"""Transform types defined in docstrings to Python parsable types.

"""

import logging
from dataclasses import dataclass, field
from pathlib import Path

import lark
import lark.visitors
from numpydoc.docscrape import NumpyDocString

from ._analysis import DocName

logger = logging.getLogger(__name__)


here = Path(__file__).parent
grammar_path = here / "doctype.lark"


def _find_one_token(tree: lark.Tree, *, name: str) -> lark.Token:
    """Find token with a specific type name in tree."""
    tokens = [child for child in tree.children if child.type == name]
    if len(tokens) != 1:
        msg = f"expected exactly one Token of type {name}, found {len(tokens)}"
        raise ValueError(msg)
    return tokens[0]


@dataclass(frozen=True, slots=True)
class PyType:
    """Python-ready type with attached import information."""

    value: str
    imports: set[DocName] | frozenset[DocName] = field(default_factory=frozenset)

    def __post_init__(self):
        object.__setattr__(self, "imports", frozenset(self.imports))

    def __str__(self) -> str:
        return self.value

    @classmethod
    def from_concatenated(cls, pytypes):
        """Concatenate multiple PyTypes in a tuple.

        Useful to combine multiple returned types for a function into a single
        PyType.

        Parameters
        ----------
        pytypes : Iterable[PyType]
            The types to combine.

        Returns
        -------
        concatenated : PyType
            The concatenated types.
        """
        values = []
        imports = set()
        for p in pytypes:
            values.append(p.value)
            imports.update(p.imports)
        value = " , ".join(values)
        if len(values) > 1:
            value = f"tuple[{value}]"
        joined = cls(value=value, imports=imports)
        return joined


class MatchedName(lark.Token):
    pass


@lark.visitors.v_args(tree=True)
class DoctypeTransformer(lark.visitors.Transformer):
    """Transformer for docstring type descriptions (doctypes).

    Parameters
    ----------
    docnames : dict[str, DocName]
        A dictionary mapping atomic names used in doctypes to information such
        as where to import from or how to replace the name itself.
    kwargs : dict[Any, Any]
        Keyword arguments passed to the init of the parent class.
    """

    def __init__(self, *, docnames, **kwargs):
        self.docnames = docnames
        self._collected_imports = None
        super().__init__(**kwargs)

    def __default__(self, data, children, meta):
        """Unpack children of rule nodes by default.

        Parameters
        ----------
        data : lark.Token
            The rule-token of the current node.
        children : list[lark.Token, ...]
            The children of the current node.
        meta : lark.tree.Meta
            Meta information for the current node.

        Returns
        -------
        out : lark.Token or list[lark.Token, ...]
            Either a token or list of tokens.
        """
        if isinstance(children, list) and len(children) == 1:
            out = children[0]
            out.type = data.upper()  # Turn rule into "token"
        else:
            out = children
        return out

    def transform(self, tree):
        """

        Parameters
        ----------
        tree : lark.Tree
            The

        Returns
        -------
        pytype : PyType
            The doctype formatted as a stub-file compatible string with
            necessary imports attached.
        """
        try:
            self._collected_imports = set()
            value = super().transform(tree=tree)
            pytype = PyType(value=value, imports=frozenset(self._collected_imports))
            return pytype
        finally:
            self._collected_imports = None

    def doctype(self, tree):
        out = " | ".join(tree.children)
        return out

    def type_or(self, tree):
        out = " | ".join(tree.children)
        return out

    def optional(self, tree):
        out = "None"
        literal = [child for child in tree.children if child.type == "LITERAL"]
        assert len(literal) <= 1
        if len(literal) == 1:
            out = lark.Discard  # Should be covered by doctype
        return out

    def extra_info(self, tree):
        logger.debug("dropping extra info")
        return lark.Discard

    def qualname(self, tree):
        matched = False
        out = []
        for i, child in enumerate(tree.children):
            if i != 0 and not child.startswith("["):
                out.append(".")
            if isinstance(child, MatchedName):
                matched = True
            out.append(child)
        out = "".join(out)
        if matched is False:
            logger.warning("unmatched name %r", out)
        return out

    def NAME(self, token):
        new_token = self._match_n_record_name(token)
        return new_token

    def ARRAY_NAME(self, token):
        new_token = self._match_n_record_name(token)
        new_token.type = "ARRAY_NAME"
        return new_token

    def shape(self, tree):
        logger.debug("dropping shape information")
        return lark.Discard

    def shape_n_dtype(self, tree):
        name = _find_one_token(tree, name="ARRAY_NAME")
        children = [child for child in tree.children if child != name]
        if children:
            name = f"{name}[{', '.join(children)}]"
        return name

    def container_of(self, tree):
        assert len(tree.children) == 2
        container_name, item_type = tree.children
        if container_name == "tuple":
            item_type += ", ..."
        out = f"{container_name}[{item_type}]"
        return out

    def contains(self, tree):
        out = ", ".join(tree.children)
        out = f"[{out}]"
        return out

    def literals(self, tree):
        out = " , ".join(tree.children)
        out = f"Literal[{out}]"
        self._collected_imports.add(self.docnames["Literal"])
        return out

    def _match_n_record_name(self, token):
        """Match type names to known imports."""
        assert "." not in token
        if token in self.docnames:
            docname = self.docnames[token]
            token = MatchedName(token.type, value=docname.use_name)
            if docname.has_import:
                self._collected_imports.add(docname)
        return token


with grammar_path.open() as file:
    _grammar = file.read()

_lark = lark.Lark(_grammar)


def doc2pytype(doctype, *, docnames):
    """Convert a type description to a Python-ready type.

    Parameters
    ----------
    doctype : str
        The type description of a parameter or return value, as extracted from
        a docstring.
    docnames : dict[str, DocName]
        A dictionary mapping atomic names used in doctypes to information such
        as where to import from or how to replace the name itself.

    Returns
    -------
    pytype : PyType
        The transformed type, ready to be inserted into a stub file, with
        necessary imports attached.
    """
    try:
        transformer = DoctypeTransformer(docnames=docnames)
        tree = _lark.parse(doctype)
        pytype = transformer.transform(tree)
        return pytype
    except Exception:
        logger.exception("couldn't parse docstring %r:", doctype)
        return PyType(
            value="Any", imports={DocName.from_cfg("Any", {"from": "typing"})}
        )


class ReturnKey:
    """Simple "singleton" key to access the return PyType in a dictionary.

    See :func:`collect_pytypes` for more.
    """


ReturnKey = ReturnKey()


def collect_pytypes(docstring, *, docnames):
    """Collect PyTypes from a docstring.

    Parameters
    ----------
    docstring : str
        The docstring to collect from.
    docnames : dict[str, DocName]
        A dictionary mapping atomic names used in doctypes to information such
        as where to import from or how to replace the name itself.

    Returns
    -------
    pytypes : dict[str | Literal[ReturnKey], PyType]
        The collected PyType for each parameter. If a return type is documented
        it's saved under the special key :class:`ReturnKey`.
    """
    np_docstring = NumpyDocString(docstring)

    params = {p.name: p for p in np_docstring["Parameters"]}
    other = {p.name: p for p in np_docstring["Other Parameters"]}
    duplicate_params = params.keys() & other.keys()
    if duplicate_params:
        raise ValueError(f"{duplicate_params=}")
    params.update(other)

    pytypes = {
        name: doc2pytype(param.type, docnames=docnames)
        for name, param in params.items()
        if param.type
    }

    returns = [
        doc2pytype(param.type, docnames=docnames)
        for param in np_docstring["Returns"]
        if param.type
    ]
    if returns:
        pytypes[ReturnKey] = PyType.from_concatenated(returns)

    return pytypes



import configparser
import logging
from collections.abc import Sequence
from typing import Any, Literal, Self, Union

logger = logging.getLogger(__name__)

__all__ = [
    "func_empty",
    "ExampleClass",
]

def func_empty(a1: Any, a2: Any, a3: Any) -> None: ...
def func_contains(
    self,
    a1: list[float],
    a2: dict[str, Union[int, str]],
    a3: Sequence[int | float],
    a4: frozenset[bytes],
) -> tuple[tuple[int, ...], list[int]]: ...
def func_literals(
    a1: Literal[1, 3, "foo"], a2: Literal["uno", 2, "drei", "four"] = ...
) -> None: ...

class ExampleClass:
    def __init__(self, a1: int, a2: float | None = ...) -> None: ...
    def method(self, a1: float, a2: float | None) -> list[float]: ...
    @staticmethod
    def some_staticmethod(a1: float, a2: float | None = ...) -> dict[str, Any]: ...
    @property
    def some_property(self) -> str: ...
    @classmethod
    def method_returning_cls(cls, config: configparser.ConfigParser) -> Self: ...



import logging
import sys
from pathlib import Path

import click

from . import _config
from ._analysis import DocName, common_docnames
from ._stubs import Py2StubTransformer, walk_python_package
from ._version import __version__

logger = logging.getLogger(__name__)


_VERBOSITY_LEVEL = {0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}


@click.command()
@click.version_option(__version__)
@click.argument("source_dir", type=click.Path(exists=True, file_okay=False))
@click.option("-o", "--out-dir", type=click.Path(file_okay=False))
@click.option("--config", "config_path", type=click.Path(exists=True, dir_okay=False))
@click.option("-v", "--verbose", count=True, help="Log more details")
@click.help_option("-h", "--help")
def main(source_dir, out_dir, config_path, verbose):
    verbose = min(2, max(0, verbose))  # Limit to range [0, 2]
    logging.basicConfig(
        level=_VERBOSITY_LEVEL[verbose],
        format="%(levelname)s: %(filename)s, line %(lineno)d, in %(funcName)s: %(message)s",
        stream=sys.stderr,
    )

    source_dir = Path(source_dir)

    # Handle configuration
    if config_path is None:
        config_path = source_dir.parent / "docstub.toml"
    else:
        config_path = Path(config_path)
    config = _config.default_config()
    if config_path.exists():
        _user_config = _config.load_config_file(config_path)
        config = _config.merge_config(config, _user_config)

    # Build docname map
    docnames = common_docnames()
    docnames.update(
        {
            name: DocName.from_cfg(docname=name, spec=spec)
            for name, spec in config["docnames"].items()
        }
    )
    # and the stub transformer
    stub_transformer = Py2StubTransformer(docnames=docnames)

    if not out_dir:
        out_dir = source_dir.parent
    out_dir = Path(out_dir) / (source_dir.name + "-stubs")
    out_dir.mkdir(parents=True, exist_ok=True)

    for source_path, stub_path in walk_python_package(source_dir, out_dir):
        if source_path.suffix.lower() == ".pyi":
            logger.debug("using existing stub file %s", source_path)
            with source_path.open() as fo:
                stub_content = fo.read()
        else:
            with source_path.open() as fo:
                py_content = fo.read()
            logger.debug("creating stub from %s", source_path)
            try:
                stub_content = stub_transformer.python_to_stub(py_content)
            except Exception as e:
                logger.exception("failed creating stub for %s:\n\n%s", source_path, e)
                continue
        stub_path.parent.mkdir(parents=True, exist_ok=True)
        with stub_path.open("w") as fo:
            logger.info("wrote %s", stub_path)
            fo.write(stub_content)



"""Basic docstring examples.

Docstrings, including module-level ones, are stripped.
"""

# Existing imports are preserved
import logging

# Assign-statements are preserved
logger = logging.getLogger(__name__)  # Inline comments are stripped


__all__ = [
    "func_empty",
    "ExampleClass",
]


def func_empty(a1, a2, a3):
    """Empty type descriptions will be replaced with "Any".

    Parameters
    ----------
    a1 :
    a2
    """


def func_contains(self, a1, a2, a3, a4):
    """Dummy.

    Parameters
    ----------
    a1 : list[float]
    a2 : dict[str, Union[int, str]]
    a3 : Sequence[int | float]
    a4 : frozenset[bytes]

    Returns
    -------
    r1 : tuple of int
    r2 : list of int
    """


def func_literals(a1, a2="uno"):
    """Dummy

    Parameters
    ----------
    a1 : {1, 3, "foo"}
    a2 : {"uno", 2, "drei", "four"}, default: "uno"
    """


class ExampleClass:
    # TODO also take into account class level docstring

    def __init__(self, a1, a2=None):
        """
        Parameters
        ----------
        a1 : int
        a2 : float, optional
        """

    def method(self, a1, a2):
        """Dummy.

        Parameters
        ----------
        a1 : float
        a2 : float, optional

        Returns
        -------
        r1 : list of float
        """

    @staticmethod
    def some_staticmethod(a1, a2="uno"):
        """Dummy

        Parameters
        ----------
        a1 : float
        a2 : float, optional

        Returns
        -------
        r1 : dict[str, Any]
        """

    @property
    def some_property(self):
        """Dummy

        Returns
        -------
        name : str
        """
        return str(self)

    @classmethod
    def method_returning_cls(cls, config):
        """Using `Self` in context of classmethods is supported.

        Parameters
        ----------
        config : configparser.ConfigParser
            Configuation.

        Returns
        -------
        out : Self
            New class.
        """



"""Collect type information."""

import builtins
import collections.abc
import typing
from dataclasses import dataclass


@dataclass(slots=True, frozen=True)
class DocName:
    """An atomic name (without ".") in a docstring type with import info."""

    use_name: str
    import_name: str = None
    import_path: str = None
    import_alias: str = None
    is_builtin: bool = False

    @classmethod
    def from_cfg(cls, docname: str, spec: dict):
        use_name = docname
        if "import" in spec:
            use_name = spec["import"]
        if "as" in spec:
            use_name = spec["as"]
        if "use" in spec:
            use_name = spec["use"]

        import_name = docname
        if "use" in spec:
            import_name = spec["use"]
        if "import" in spec:
            import_name = spec["import"]

        docname = cls(
            use_name=use_name,
            import_name=import_name,
            import_path=spec.get("from"),
            import_alias=spec.get("as"),
            is_builtin=spec.get("builtin", False),
        )
        return docname

    def format_import(self):
        if self.is_builtin:
            msg = "cannot import builtin"
            raise RuntimeError(msg)
        out = f"import {self.import_name}"
        if self.import_path:
            out = f"from {self.import_path} {out}"
        if self.import_alias:
            out = f"{out} as {self.import_alias}"
        return out

    @property
    def has_import(self):
        return not self.is_builtin


def _is_type(value) -> bool:
    """Check if value is a type."""
    # Checking for isinstance(..., type) isn't enough, some types such as
    # typing.Literal don't pass that check. So combine with checking for a
    # __class__ attribute. Not sure about edge cases!
    is_type = isinstance(value, type) or hasattr(value, "__class__")
    return is_type


def _builtin_docnames():
    """Return docnames for all builtins (in the current runtime).

    Returns
    -------
    docnames : dict[str, DocName]
    """
    known_builtins = set(dir(builtins))

    docnames = {}
    for name in known_builtins:
        if name.startswith("_"):
            continue
        value = getattr(builtins, name)
        if not _is_type(value):
            continue
        docnames[name] = DocName(use_name=name, is_builtin=True)

    return docnames


def _typing_docnames():
    """Return docnames for public types in the `typing` module.

    Returns
    -------
    docnames : dict[str, DocName]
    """
    docnames = {}
    for name in typing.__all__:
        if name.startswith("_"):
            continue
        value = getattr(typing, name)
        if not _is_type(value):
            continue
        docnames[name] = DocName.from_cfg(name, spec={"from": "typing"})
    return docnames


def _collections_abc_docnames():
    """Return docnames for public types in the `collections.abc` module.

    Returns
    -------
    docnames : dict[str, DocName]
    """
    docnames = {}
    for name in collections.abc.__all__:
        if name.startswith("_"):
            continue
        value = getattr(collections.abc, name)
        if not _is_type(value):
            continue
        docnames[name] = DocName.from_cfg(name, spec={"from": "collections.abc"})
    return docnames


def common_docnames():
    """Return docnames for commonly supported types.

    This includes builtin types, and types from the `typing` or
    `collections.abc` module.

    Returns
    -------
    docnames : dict[str, DocName]
    """
    docnames = _builtin_docnames()
    docnames |= _typing_docnames()
    docnames |= _collections_abc_docnames()  # Overrides containers from typing
    return docnames


def find_module_paths_using_search():
    # TODO use from mypy.stubgen import find_module_paths_using_search ?
    #   https://github.com/python/mypy/blob/66b48cbe97bf9c7660525766afe6d7089a984769/mypy/stubgen.py#L1526
    pass



import numpy as np
from numpy.typing import ArrayLike, NDArray

def func_object_with_numpy_objects(
    a1: np.np.int8, a2: np.np.int16, a3: np.typing.DTypeLike, a4: np.typing.DTypeLike
) -> None: ...
def func_ndarray(
    a1: NDArray, a2: np.NDArray, a3: NDArray[float], a4: NDArray[np.uint8] = ...
) -> tuple[NDArray[np.uint8], NDArray[complex]]: ...
def func_array_like(
    a1: ArrayLike, a2: ArrayLike, a3: ArrayLike[float], a4: ArrayLike[np.uint8]
) -> tuple[ArrayLike[np.uint8], ArrayLike[complex]]: ...



import logging
from pathlib import Path

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib


logger = logging.getLogger(__name__)


DEFAULT_CONFIG_PATH = Path(__file__).parent / "default_config.toml"


def load_config_file(path: Path | str) -> dict:
    """Return configuration options in local TOML file if they exist."""
    with open(path, "rb") as fp:
        config = tomllib.load(fp)
    config = config.get("tool", {}).get("docstub", {})
    return config


def default_config():
    config = load_config_file(DEFAULT_CONFIG_PATH)
    return config


def merge_config(*configurations):
    merged = {}
    merged["extended_grammar"] = "\n".join(
        cfg.get("extended_grammar", "") for cfg in configurations
    )
    merged["docnames"] = {}
    for cfg in configurations:
        docnames = cfg.get("docnames")
        if docnames and isinstance(docnames, dict):
            merged["docnames"].update(docnames)
    return merged



"""NumPy and array specific docstring examples."""


def func_object_with_numpy_objects(a1, a2, a3, a4):
    """Dummy.

    Parameters
    ----------
    a1 : numpy.int8
    a2 : np.int16
    a3 : numpy.typing.DTypeLike
    a4 : np.typing.DTypeLike
    """


def func_ndarray(a1, a2, a3, a4=None):
    """Dummy.

    Parameters
    ----------
    a1 : ndarray
    a2 : np.ndarray
    a3 : (N, 3) ndarray of float
    a4 : ndarray of shape (1,) and dtype uint8

    Returns
    -------
    r1 : uint8 array
    r2 : array of dtype complex and shape (1, ..., 3)
    """


def func_array_like(a1, a2, a3, a4):
    """Dummy.

    Parameters
    ----------
    a1 : array-like
    a2 : array_like
    a3 : (N, 3) array-like of float
    a4 : array-like of shape (1,) and dtype uint8

    Returns
    -------
    r1 : uint8 array-like
    r2 : array_like of dtype complex and shape (1, ..., 3)
    """



"""Transform Python source files to typed stub files.

"""

import logging
from dataclasses import dataclass
from typing import Literal

import libcst as cst

from ._docstrings import ReturnKey, collect_pytypes

logger = logging.getLogger(__name__)


def walk_python_package(root_dir, target_dir):
    """Iterate modules in a Python package and it's target stub files.

    Parameters
    ----------
    root_dir : Path
        Root directory of a Python package.
    target_dir : Path
        Root directory in which a matching stub package will be created.

    Returns
    -------
    source_path : Path
        Either a Python file or a stub file that takes precedence.
    stub_path : Path
        Target stub file.

    Notes
    -----
    Files starting with "test_" are skipped entirely for now.
    """
    for root, _, files in root_dir.walk(top_down=True):
        for name in files:
            source_path = root / name

            if name.startswith("test_"):
                logger.debug("skipping %s", name)
                continue

            if source_path.suffix.lower() not in {".py", ".pyi"}:
                continue

            if (
                source_path.suffix.lower() == ".py"
                and source_path.with_suffix(".pyi").exists()
            ):
                # Stub file already exists and takes precedence
                continue

            stub_path = target_dir / source_path.with_suffix(".pyi").relative_to(
                root_dir
            )
            yield source_path, stub_path


def try_format_stub(stub: str) -> str:
    """Try to format a stub file with isort and black if available."""
    try:
        import isort

        stub = isort.code(stub)
    except ImportError:
        logger.warning("isort is not available, couldn't sort imports")
    try:
        import black

        stub = black.format_str(stub, mode=black.Mode(is_pyi=True))
    except ImportError:
        logger.warning("black is not available, couldn't format stubs")
    return stub


@dataclass(slots=True, frozen=True)
class _Scope:
    type: Literal["module", "class", "func", "method", "classmethod", "staticmethod"]
    node: cst.CSTNode = None

    def __post_init__(self):
        allowed_types = {
            "module",
            "class",
            "func",
            "method",
            "classmethod",
            "staticmethod",
        }
        if self.type not in allowed_types:
            msg = f"type {self.type!r} is not allowed, allowed are {allowed_types!r}"
            raise ValueError(msg)

    @property
    def has_self_or_cls(self):
        return self.type in {"method", "classmethod"}


class Py2StubTransformer(cst.CSTTransformer):
    """Transform syntax tree of a Python file into the tree of a stub file."""

    # Equivalent to ` ...`, to replace the body of callables with
    _body_replacement = cst.SimpleStatementSuite(
        leading_whitespace=cst.SimpleWhitespace(value=" "),
        body=[cst.Expr(value=cst.Ellipsis())],
    )
    _Annotation_Any = cst.Annotation(cst.Name("Any"))
    _Annotation_None = cst.Annotation(cst.Name("None"))

    def __init__(self, *, docnames):
        self.docnames = docnames
        # Relevant docstring for the current context
        self._scope_stack = None  # Store current class or function scope
        self._pytypes_stack = None  # Store current parameter types
        self._required_imports = None  # Collect imports for used types

    def python_to_stub(self, source: str) -> str:
        """Convert Python source code to stub-file ready code."""
        try:
            self._scope_stack = []
            self._pytypes_stack = []
            self._required_imports = set()

            source_tree = cst.parse_module(source)
            stub_tree = source_tree.visit(self)
            stub = stub_tree.code
            stub = try_format_stub(stub)
            return stub
        finally:
            assert len(self._scope_stack) == 0
            assert len(self._pytypes_stack) == 0
            self._scope_stack = None
            self._pytypes_stack = None
            self._required_imports = None

    def visit_ClassDef(self, node):
        self._scope_stack.append(_Scope(type="class", node=node))
        return True

    def leave_ClassDef(self, original_node, updated_node):
        self._scope_stack.pop()
        return updated_node

    def visit_FunctionDef(self, node):
        func_type = "func"
        if self._scope_stack[-1].type == "class":
            func_type = "method"
        for decorator in node.decorators:
            assert func_type in {"func", "method"}
            if decorator.decorator.value == "classmethod":
                func_type = "classmethod"
                break
            if decorator.decorator.value == "staticmethod":
                func_type = "staticmethod"
                break
        self._scope_stack.append(_Scope(type=func_type, node=node))

        docstring = node.get_docstring()
        pytypes = None
        if docstring:
            try:
                pytypes = collect_pytypes(docstring, docnames=self.docnames)
            except Exception as e:
                logger.exception(
                    "error while parsing docstring of `%s`:\n\n%s", node.name.value, e
                )
        self._pytypes_stack.append(pytypes)
        return True

    def leave_FunctionDef(self, original_node, updated_node):
        node_changes = {
            "body": self._body_replacement,
            "returns": self._Annotation_None,
        }

        pytypes = self._pytypes_stack.pop()
        if pytypes:
            return_pytype = pytypes.get(ReturnKey)
            if return_pytype:
                node_changes["returns"] = cst.Annotation(
                    cst.parse_expression(return_pytype.value)
                )
                self._required_imports |= return_pytype.imports

        updated_node = updated_node.with_changes(**node_changes)
        self._scope_stack.pop()
        return updated_node

    def leave_Param(self, original_node, updated_node):
        node_changes = {}

        scope = self._scope_stack[-1]
        # Check if is first parameter of method or classmethod
        is_self_or_cls = (
            scope.node.params.children[0] is original_node and scope.has_self_or_cls
        )

        name = original_node.name.value
        pytypes = self._pytypes_stack[-1]
        if pytypes:
            pytype = pytypes.get(name)
            if pytype:
                annotation = cst.Annotation(cst.parse_expression(pytype.value))
                node_changes["annotation"] = annotation
                if pytype.imports:
                    self._required_imports |= pytype.imports

        # Potentially use "Any" except for first param in (class)methods
        elif not is_self_or_cls and updated_node.annotation is None:
            node_changes["annotation"] = self._Annotation_Any
            self._required_imports.add(self.docnames["Any"])

        if updated_node.default is not None:
            node_changes["default"] = cst.Ellipsis()

        if node_changes:
            updated_node = updated_node.with_changes(**node_changes)
        return updated_node

    def leave_Expr(self, original_node, upated_node):
        return cst.RemovalSentinel.REMOVE

    def leave_Comment(self, original_node, updated_node):
        return cst.RemovalSentinel.REMOVE

    def visit_Module(self, node):
        self._scope_stack.append(_Scope(type="module", node=node))
        return True

    def leave_Module(self, original_node, updated_node):
        import_nodes = self._parse_imports(self._required_imports)
        updated_node = updated_node.with_changes(body=import_nodes + updated_node.body)
        self._scope_stack.pop()
        return updated_node

    @staticmethod
    def _parse_imports(imports):
        """Create nodes to include in the module tree from given imports.

        Parameters
        ----------
        imports : set[DocName]

        Returns
        -------
        import_nodes : tuple[cst.SimpleStatementLine, ...]
        """
        lines = {imp.format_import() for imp in imports}
        import_nodes = tuple(cst.parse_statement(line) for line in lines)
        return import_nodes
